#include <stdio.h>
#include <string.h>

#define FAILED(x) (x<0)

typedef struct CLex {
  FILE *fp;
  int ch;
} CLex ;

static int CLex_Open(CLex *__this, char *filename) {
  if((__this->fp=fopen(filename,"r"))==NULL)
    return -1;
  __this->ch = fgetc(__this->fp);
  return 0;
}

static int CLex_Close(CLex *__this) {
  fclose(__this->fp);
  return 0;
}

static int CLex_SkipLine(CLex *__this) {
  while(__this->ch!='\n'&&__this->ch!=EOF)
    __this->ch=fgetc(__this->fp);
  return 0;
}

static int CLex_SkipSpace(CLex *__this) {
  while(strchr(" \t\n,;",__this->ch)) {
    if (__this->ch==';')
      CLex_SkipLine(__this);
    else
      __this->ch=fgetc(__this->fp);
  }
  return 0;
}

static int CLex_GetNumber(CLex *__this) {
  int ret=-1;
  CLex_SkipSpace(__this);
  while( '0'<=__this->ch && __this->ch <= '9' ) {
    ret = (ret<0?0:ret) * 10 + (__this->ch - '0');
    __this->ch=fgetc(__this->fp); 
  }
  return ret;
}

static int CLex_SkipDollerSpace(CLex *__this) {
  while(strchr("$ \t\n,;",__this->ch)) {
    if (__this->ch==';')
      CLex_SkipLine(__this);
    else
      __this->ch=fgetc(__this->fp);
  }
  return 0;
}

static int CLex_GetHexNumber(CLex *__this) {
  int ret=-1;
  CLex_SkipDollerSpace(__this);
  while( ('0'<=__this->ch && __this->ch <= '9') || ('A'<=__this->ch && __this->ch <= 'F')  || ('a'<=__this->ch && __this->ch <= 'f') ) {
    int num;
    if ('0'<=__this->ch && __this->ch <= '9'){
      num = __this->ch - '0';
    }else if('A'<=__this->ch && __this->ch <= 'F'){
      num = 10 + __this->ch - 'A';
    }else if('a'<=__this->ch && __this->ch <= 'f'){
      num = 10 + __this->ch - 'a';
    }
    //printf("num%d,$%x\n",num,num);	//debug
    ret = (ret<0?0:ret) * 16 + num;
    __this->ch=fgetc(__this->fp); 
  }
  return ret;
}

typedef struct TONE
{
  int tl, fb, ml, kl, ar, dr, sl, rr, am, pm, eg, kr, wf;
} TONE;

static int tone2hex(TONE t[2], int hex[8]) {
  int i;
  for(i=0;i<2;i++) {
    hex[0+i] = (t[i].am<<7) | (t[i].pm<<6) | 
               (t[i].eg<<5) | (t[i].kr<<4) | (t[i].ml);
    hex[2+i] = (t[i].kl<<6);
    hex[4+i] = (t[i].ar<<4) | (t[i].dr);
    hex[6+i] = (t[i].sl<<4) | (t[i].rr);
  }
  hex[2] |= t[0].tl;
  hex[3] |= (t[1].wf<<4)|(t[0].wf<<3)|(t[0].fb);
  return 0;
}

static int txt2tone(char *filename, TONE t[2]) {

  CLex lex;
 
  int i;

  if(FAILED(CLex_Open(&lex,filename)))
    return -1;

  t[0].tl = CLex_GetNumber(&lex);
  if(t[0].tl<0||63<t[0].tl) return -1;
  t[0].fb = CLex_GetNumber(&lex);
  if(t[0].fb<0||7<t[0].fb) return -1;
  for(i=0;i<2;i++) {
    t[i].ar = CLex_GetNumber(&lex);
    if(t[i].ar<0||15<t[i].ar) return -1;
    t[i].dr = CLex_GetNumber(&lex);
    if(t[i].dr<0||15<t[i].dr) return -1;
    t[i].sl = CLex_GetNumber(&lex);
    if(t[i].sl<0||15<t[i].sl) return -1;
    t[i].rr = CLex_GetNumber(&lex);
    if(t[i].rr<0||15<t[i].rr) return -1;
    t[i].kl = CLex_GetNumber(&lex);
     if(t[i].kl<0||3<t[i].kl) return -1;
    t[i].ml = CLex_GetNumber(&lex);
    if(t[i].ml<0||15<t[i].ml) return -1;
    t[i].am = CLex_GetNumber(&lex);
    if(t[i].am<0||1<t[i].am) return -1;
    t[i].pm = CLex_GetNumber(&lex);
    if(t[i].pm<0||1<t[i].pm) return -1;
    t[i].eg = CLex_GetNumber(&lex);
    if(t[i].eg<0||1<t[i].eg) return -1;
    t[i].kr = CLex_GetNumber(&lex);
    if(t[i].kr<0||1<t[i].kr) return -1;
    t[i].wf = CLex_GetNumber(&lex);
    if(t[i].wf<0||1<t[i].wf) return -1;
  }

  CLex_Close(&lex);
  return 0;
}

void help(){
  fprintf(stderr,"Usage: oplltone [options] textfile\n");
  fprintf(stderr,"(Options)\n");
  fprintf(stderr,"  /y : output y command sequence.\n");
  fprintf(stderr,"  /r : reverse convert\n");
}

static int txt2hex(char *filename, int hex[8]) {

  CLex lex;
 
  int i;

  if(FAILED(CLex_Open(&lex,filename)))
    return -1;

  for (i=0;i<8;i++){
     hex[i] = CLex_GetHexNumber(&lex);
     //printf("%d($%x)\n",hex[i],hex[i]);	//debug
  }

  CLex_Close(&lex);
  return 0;
}

static int hex2tone(int hex[8], TONE t[2]) {
  int i;
  for(i=0;i<2;i++) {
    t[i].am = (hex[0+i]>>7) &  1;
    t[i].pm = (hex[0+i]>>6) &  1;
    t[i].eg = (hex[0+i]>>5) &  1;
    t[i].kr = (hex[0+i]>>4) &  1;
    t[i].ml = (hex[0+i]   ) & 15;
    t[i].kl = (hex[2+i]>>6) &  3;
    t[i].ar = (hex[4+i]>>4) & 15;
    t[i].dr = (hex[4+i]   ) & 15;
    t[i].sl = (hex[6+i]>>4) & 15;
    t[i].rr = (hex[6+i]   ) & 15;
  }
  t[0].tl = (hex[2]   ) & 63;
  t[0].wf = (hex[3]>>3) &  1;
  t[1].wf = (hex[3]>>4) &  1;
  t[0].fb = (hex[3]   ) &  7;
  return 0;
}

void printTone(TONE t[2]){
    int i;
    printf(";TL FB\n");
    printf(" %02d %02d\n", t[0].tl, t[0].fb);
    printf(";AR DR SL RR KL ML AM PM EG KR WF\n");
    for(i=0;i<2;i++)
      printf(" %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d %2d\n", 
             t[i].ar,t[i].dr,t[i].sl,t[i].rr,t[i].kl,t[i].ml,
             t[i].am,t[i].pm,t[i].eg,t[i].kr,t[i].wf);
}

int main(int argc, char *argv[])
{
  TONE t[2];
  int v[8];
  int i;
  char *input;
  int opt_ycmd=0;
  int opt_reverse=0;

  if(argc<2){
    help();
    return 0;
  }

  for(i=1;i<argc;i++) {
    if(argv[i][0]=='-'||argv[i][0]=='/') {
      switch(argv[i][1]) {
      case 'y':
        opt_ycmd = 1;
        break;
      case 'r':
        opt_reverse=1;
        break;
      default:
        help();
        return 1; 
      }
    }
    else input = argv[i];
  }

  memset(&t,-1,sizeof(t));

  if (opt_reverse){
    if(FAILED(txt2hex(input, v))) {
      fprintf(stderr,"Error.\n");
      for(i=0;i<8;i++)
        printf("$%02x ",v[i]);
    }else{
      hex2tone(v, t);
      printTone(t);
    }
  }else if(FAILED(txt2tone(input, t))) {
    fprintf(stderr,"Error.\n");
    printTone(t);
  } else {
    tone2hex(t,v);
    for(i=0;i<8;i++)
      if(opt_ycmd)
        printf("y$9010,$%02xy$9030,$%02x\n",i,v[i]);
      else
        printf("$%02x ",v[i]);
    printf("\n"); 
  }

  return 0;
}
